from time import time

import numpy as np
from pandas import DataFrame

from .HandData.OneHand import OneHand
from .HandData.Gestrue import Gestrue
from .HandData._global import HAND_DATA_COL_NAME
from .HandDetector.VisualHandDetector import VisualHandDetector
from .HandDetector.MediaPipeHandDetector import MediaPipeHandDetector
from .HandDetector.HandsMatchers import HandsMatcher, HungarianMatcher
from .HandUtils.Drawing import HandDrawing
from .HandInputSchemes.HandInputScheme import HandInputScheme


class BaseHandAPI:
    __slots__ = "one_hand", "delta_time"

    def __init__(self, one_hand: OneHand) -> None:
        self.one_hand: OneHand = one_hand
        # 记录帧间时间,初始值为1
        self.delta_time: float = 1

    def hand_side(self) -> str:
        """获取该手部是左手还是右手"""
        return self.one_hand.hand_side

    def img_pos(self, point_id: int = -1) -> np.ndarray:
        """获取该手部的第point_id个关键点在图片中的像素坐标
        Args:
            point_id: 关键点编号;默认值-1返回所有关键点的像素坐标
        """
        if 20 >= point_id >= 0:  # 返回指定关键点的像素坐标的副本
            return self.one_hand.raw_pos[point_id, :2].astype(np.int32)
        elif point_id == -1:  # 返回所有关键点的像素坐标的副本
            return self.one_hand.raw_pos[:, :2].astype(np.int32)
        raise ValueError(f"No coordinate data with point_id {point_id}")

    def norm_pos(self, point_id: int = -1, copy: bool = True) -> np.ndarray:
        """获取该手部的索引为point_id的以手部矩形框左上角为原点的归一化xyz坐标
        Args:
            point_id: 关键点编号;默认值-1返回所有关键点的归一化坐标
            copy: 是否返回副本,默认为True返回副本,注意:若为False,请不要修改该值,这是只读的
        """
        if 20 >= point_id >= 0:  # 返回指定的关键点归一化坐标
            output = self.one_hand.norm_pos[point_id, :]
        elif point_id == -1:  # 返回该手部所有的关键点的归一化坐标
            output = self.one_hand.norm_pos
        else:
            raise ValueError(f"There is no coordinate data with point_id {point_id}")
        # 返回nrom_pos的副本,防止原来的数据被修改
        return output.copy() if copy else output

    def wrist_npos(self, point_id: int = -1, copy: bool = True) -> np.ndarray:
        """获取该手部的索引为point_id的以手腕为原点的归一化xyz坐标
        Args:
            point_id: 关键点编号;默认值-1返回所有关键点的以手腕为原点的归一化坐标
            copy: 是否返回副本,默认为True返回副本,注意:若为False,请不要修改该值,这是只读的
        """
        if 20 >= point_id > 0:  # 返回指定的关键点归一化坐标
            output = self.one_hand.wrist_npos[point_id, :]
        elif point_id == -1:  # 返回该手部所有的关键点的归一化坐标
            output = self.one_hand.wrist_npos
        elif point_id == 0:  # 以手腕坐标为原点,直接返回原点坐标000
            return np.array([0, 0, 0], dtype=np.float32)
        else:
            raise ValueError(f"There is no coordinate data with point_id {point_id}")
        return output.copy() if copy else output

    def norm2img_pos(
        self,
        point_id: int = -1,
        img_size: tuple[int, int] = (100, 100),
        padding: tuple[int, int] = (0, 0),
    ) -> np.ndarray:
        """将归一化后的坐标转换为特定大小的图片位置坐标,返回坐标数组
        Args:
            point_id: 该手部的关键点编号
            img_size: 转换的图片尺寸,(图片宽,图片高)
            padding: 图片周围的间隔距离,(padx,pady)
        """
        if point_id > 20 or point_id < -1:
            raise ValueError(f"There is no coordinate data with point_id {point_id}")
        # 获取对应手部的关键点坐标
        npos = (
            self.one_hand.norm_pos[point_id, :2]
            if 20 >= point_id >= 0
            else self.one_hand.norm_pos[:, :2]
        )
        img_size_arr = np.array(img_size[::-1])
        padding_arr = np.array(padding)
        # 计算转换后的图片坐标
        ipos = (npos * (img_size_arr - (2 * padding_arr))) + padding_arr
        return ipos.astype(np.int32)

    def box(self) -> np.ndarray:
        """获取对应名字的手部最小矩形框的在图片中的xyxy像素坐标"""
        return self.one_hand.box

    def finger_angle(self, point_id: int = -1, copy: bool = True) -> np.ndarray:
        """获取对应名字的手部关键点弧度制角度,其中指尖和手腕没有角度数据
        Args:
            point_id: 关键点编号;默认值-1返回所有手指关节点角度数据
            copy: 是否返回副本,默认为True返回副本,注意:若为False,请不要修改该值,这是只读的
        """
        finger_id, angle_id = divmod(point_id, 4)  # 计算关键点对应的角度数组中的索引
        if 20 >= point_id > 0 and angle_id != 0:  # 指尖和手腕没有角度数据
            output = self.one_hand.fingers_angle[finger_id, (angle_id - 1)]
        elif point_id == -1:  # 没有输入point_id则返回全部角度
            output = self.one_hand.fingers_angle
        else:
            raise ValueError(f"There is no angle data with point_id {point_id}")
        return output.copy() if copy else output

    def thumb_dist(self, other_point_id: int = -1, copy: bool = True) -> np.ndarray:
        """获取对应名字的手部的从大拇指到其他手指关键点的曼哈顿距离
        Args:
            point_id: 关键点编号;默认值-1返回4根指尖到大拇指指尖的距离
            copy: 是否返回副本,默认为True返回副本,注意:若为False,请不要修改该值,这是只读的
        """
        if other_point_id in (8, 12, 16, 20):
            # 计算对应的数组的索引
            finger_id, knuckle_id = divmod(other_point_id, 4)
            arr_id = (knuckle_id - 1) if knuckle_id != 0 else (finger_id + 1)
            output = self.one_hand.thumb_dist[arr_id]
        elif other_point_id == -1:  # 没有输入id则返回全部到拇指的距离
            output = self.one_hand.thumb_dist
        else:
            raise ValueError(
                f"There is no distance data with point_id {other_point_id}"
            )
        return output.copy() if copy else output

    def wrist_npos_velocity(self, point_id: int = -1) -> np.ndarray:
        """获取该手部的第point_id个关键点两帧之间的归一化坐标差值
        Args:
            point_id: 关键点编号,默认值-1则返回所有两帧差坐标
        """
        if 20 >= point_id >= 0:  # 返回指定关键点的两帧像素差坐标
            return self.one_hand.wrist_npos_diff[point_id] / self.delta_time
        elif point_id == -1:  # 返回所有关键点的两帧像素差坐标
            return self.one_hand.wrist_npos_diff / self.delta_time
        raise ValueError(f"There is no coordinate data with point_id {point_id}")

    def angle_velocity(self, point_id: int = -1) -> np.ndarray:
        """获取对应名字的手部关键点弧度制角度差,其中指尖和手腕没有角度数据
        Args:
            point_id: 关键点编号,默认值-1则返回所有角度差
        """
        finger_id, angle_id = divmod(point_id, 4)  # 计算关键点对应的角度数组中的索引
        if 20 >= point_id > 0 and angle_id != 0:  # 指尖和手腕没有角度数据
            return self.one_hand.angle_diff[finger_id, (angle_id - 1)] / self.delta_time
        elif point_id == -1:  # 没有输入point_id则返回全部角度差
            return self.one_hand.angle_diff / self.delta_time
        raise ValueError(f"There is no angle data with point_id {point_id}")


class HandDataAPI:
    __slots__ = "one_hand"

    def __init__(self, one_hand: OneHand) -> None:
        self.one_hand: OneHand = one_hand

    @property
    def norm_pos(self) -> np.ndarray:
        """获取对应名字的手部的所有连续型数据的数组"""
        return self.one_hand.pos_data

    @property
    def norm_pos2DataFrame(self) -> DataFrame:
        """获取对应名字的手部的所有连续型数据的DataFrame格式的数据"""
        return DataFrame(self.one_hand.pos_data, columns=HAND_DATA_COL_NAME.iloc[:63])

    @property
    def finger(self) -> np.ndarray:
        """获取对应名字的手部的手指角度和指尖距离数据的数组"""
        return self.one_hand.finger_data

    @property
    def finger2DataFrame(self) -> DataFrame:
        """获取对应名字的手部的手指角度和指尖距离数据的DataFrame格式的数据"""
        return DataFrame(
            self.one_hand.finger_data, columns=HAND_DATA_COL_NAME.iloc[63:]
        )

    @property
    def all_data(self) -> np.ndarray:
        """获取对应名字的手部的所有连续型数据的数组"""
        return self.one_hand.data

    @property
    def all2DataFrame(self) -> DataFrame:
        """获取对应名字的手部的所有连续型数据的DataFrame格式的数据"""
        return DataFrame(self.one_hand.data, columns=HAND_DATA_COL_NAME)


class HandAPI:
    __slots__ = "one_hand", "_base", "_data", "_gestrue", "_drawing"

    def __init__(self, one_hand: OneHand):
        """用于管理一只手部的所有相关数据和模型
        Args:
            one_hand: 输入该手部的基础数据类实例
        """
        self.one_hand = one_hand
        self._base: BaseHandAPI = BaseHandAPI(self.one_hand)
        self._data: HandDataAPI = HandDataAPI(self.one_hand)
        self._gestrue: Gestrue = Gestrue(self.one_hand)
        self._drawing: HandDrawing = HandDrawing(self.one_hand)

    @property
    def base(self) -> BaseHandAPI:
        """返回该手部的基本手部数据的API"""
        return self._base

    @property
    def data(self) -> HandDataAPI:
        """返回该手部的一维数据格式的手部数据的API"""
        return self._data

    @property
    def gestrue(self) -> Gestrue:
        """返回该手部的手指状态相关数据的API"""
        return self._gestrue

    @property
    def drawing(self) -> HandDrawing:
        """返回绘制手部相关的API"""
        return self._drawing


class HandInput:
    __slots__ = (
        "hands_dict",
        "one_hand_dict",
        "detector",
        "detected_name_ls",
        "frame",
        "_last_frame_time",
        "_delta_time",
        "schemes",
    )

    def __init__(
        self,
        hands_name_ls: list[str],
        hands_detector: type[VisualHandDetector] = MediaPipeHandDetector,
        hands_matcher: type[HandsMatcher] = HungarianMatcher,
        **detector_kwargs,
    ) -> None:
        # 利用传入的手部名字来创建对应的手部API字典
        self.hands_dict: dict[str, HandAPI] = {
            name: HandAPI(OneHand()) for name in hands_name_ls
        }
        self.one_hand_dict: dict[str, OneHand] = {  # 手部数据字典用作手部检测的参数
            name: hand.one_hand for name, hand in self.hands_dict.items()
        }
        # 根据传入的检测器类创建视觉手部检测器实例,添加手部名字列表的数量参数
        detector_kwargs["hands_name_ls"] = hands_name_ls
        detector_kwargs["hands_matcher"] = hands_matcher
        self.detector: VisualHandDetector = hands_detector(**detector_kwargs)
        # 创建一个用于记录本次检测到的手部的名称的列表
        self.detected_name_ls: list[str] = []
        # 记录每帧的图像
        self.frame: None | np.ndarray = None
        # 记录两帧之间的间隔时间,初始值为1
        self._last_frame_time: float = time()
        self._delta_time: float = 1
        # 创建存放手部操控方案的字典
        self.schemes: dict[str, HandInputScheme] = dict()

    def run(self, frame: np.ndarray) -> list[str]:
        """运行手部关键点检测器,返回成功检测到的手部名称
        Args:
            frame: 输入需要检测的帧图像
        """
        self.frame = frame  # 记录帧图像变量
        # 检测并更新手部数据
        self.detected_name_ls = self.detector.detect(frame, self.one_hand_dict)
        # 更新帧间时间
        cur_frame_time = time()
        self._delta_time = cur_frame_time - self._last_frame_time
        self._last_frame_time = cur_frame_time
        # 实时更新手部操控方案
        for scheme in self.schemes.values():
            scheme.update()
        return self.detected_name_ls

    @property
    def delta_time(self) -> float:
        """返回两帧之间的间隔时间,初始值为1"""
        return self._delta_time

    def hand(self, hand_name: str) -> HandAPI | None:
        """指定相应的手部,返回对应手部的API,没有检测到的返回None
        Args:
            hand_name: 当前检测到的手部名字,没有检测到则返回None
        """
        # 检查指定的手部是否有检测到
        if hand_name not in self.detected_name_ls:
            return None
        return self.hands_dict[hand_name]

    def base(self, hand_name: str) -> BaseHandAPI | None:
        """获取基础的手部数据,没有检测到的返回None
        Args:
            hand_name: 当前检测到的手部名字,没有检测到则返回None
        """
        hand = self.hand(hand_name)
        if hand is None:
            return None
        hand.base.delta_time = self.delta_time
        return hand.base

    def data(self, hand_name: str) -> HandDataAPI | None:
        """获取一维的手部相关数据,没有检测到的返回None
        Args:
            hand_name: 当前检测到的手部名字,没有检测到则返回None
        """
        hand = self.hand(hand_name)
        if hand is None:
            return None
        return hand.data

    def gestrue(self, hand_name: str) -> Gestrue | None:
        """获取手势相关数据,没有检测到的返回None
        Args:
            hand_name: 当前检测到的手部名字,没有检测到则返回None
        """
        hand = self.hand(hand_name)
        if hand is None:
            return None
        return hand.gestrue

    def drawing(self, hand_name: str) -> HandDrawing | None:
        """获取手部绘制工具,没有检测到的返回None
        Args:
            hand_name: 当前检测到的手部名字,没有检测到则返回None
        """
        hand = self.hand(hand_name)
        if hand is None:
            return None
        hand.drawing.raw_img = self.frame  # 设定原始图像
        return hand.drawing

    def hand_unwrap(self, hand_name: str) -> HandAPI:
        """指定相应的手部,返回对应手部的API,没有检测到就抛出错误
        Args:
            hand_name: 当前检测到的手部名字,没有检测到则报错
        """
        # 检查指定的手部是否有检测到
        if hand_name not in self.detected_name_ls:
            raise ValueError(f"No hand named '{hand_name}' was detected")
        return self.hands_dict[hand_name]

    def base_unwrap(self, hand_name: str) -> BaseHandAPI:
        """获取基础的手部数据,没有检测到就抛出错误
        Args:
            hand_name: 当前检测到的手部名字,没有检测到则报错
        """
        hand = self.hand_unwrap(hand_name)
        hand.base.delta_time = self.delta_time  # 设定时间间隔
        return hand.base

    def data_unwrap(self, hand_name: str) -> HandDataAPI:
        """获取一维的手部相关数据,没有检测到就抛出错误
        Args:
            hand_name: 当前检测到的手部名字,没有检测到则报错
        """
        return self.hand_unwrap(hand_name).data

    def gestrue_unwrap(self, hand_name: str) -> Gestrue:
        """获取手势相关数据,没有检测到就抛出错误
        Args:
            hand_name: 当前检测到的手部名字,没有检测到则报错
        """
        return self.hand_unwrap(hand_name).gestrue

    def drawing_unwrap(self, hand_name: str) -> HandDrawing:
        """获取手部绘制工具,没有检测到就抛出错误
        Args:
            hand_name: 当前检测到的手部名字,没有检测到则报错
        """
        hand = self.hand_unwrap(hand_name)
        hand.drawing.raw_img = self.frame  # 设定原始图像
        return hand.drawing

    def add_scheme(
        self, hand_name: str, scheme_name: str, new_scheme: type[HandInputScheme]
    ):
        """用默认参数来添加新的手部操控方案
        Args:
            hand_name: 当前检测到的手部名字,没有该手部名字则报错
            scheme_name: 设置新增的手部操控方案的名字
            scheme: 输入手部操控方案实例,创建时需设置好参数
        """
        if hand_name not in self.hands_dict.keys():
            raise ValueError(f"No hand named '{hand_name}' is not exist")
        # 使用默认参数来添加新的手部操控方案
        self.schemes[scheme_name] = new_scheme(self, hand_name)



from abc import ABC, abstractmethod
from typing import Any


class HandInputScheme(ABC):
    __slots__ = "hand_input", "hand_name"

    @abstractmethod
    def __init__(self, hand_input: "HandInput", hand_name: str) -> None:
        """手部操控方案抽象类,通过调用手部数据自定义通用的手部操控方案
        Args:
            hand_input: 手部输入类的实例
            hand_name: 手部名字,指定为哪只手制定手部操控方案
        """
        hand_input.hands_dict[hand_name]  # 获取该手部,没有该手部名字则报错
        self.hand_input = hand_input
        self.hand_name: str = hand_name

    @abstractmethod
    def update(self) -> Any:
        """实时更新手部操控方案,外部通过调用其他类属性或方法来使用手部操控方案"""
        pass

    @property
    @abstractmethod
    def is_activate(self) -> bool:
        """检查该手部操控方案是否有激活,激活则返回True"""
        return True



import numpy as np
import cv2
import mediapipe as mp

from .VisualHandDetector import VisualHandDetector
from .HandsMatchers import HandsMatcher
from ..HandData.OneHand import OneHand


class MediaPipeHandDetector(VisualHandDetector):
    __slots__ = "hands_name_ls", "_detector", "hands_matcher", "matched_counter"

    def __init__(
        self,
        hands_name_ls: list[str],
        hands_matcher: type[HandsMatcher],
        *,
        static_image_mode: bool = False,
        min_detect_confi: float = 0.8,
        min_track_confi: float = 0.6,
    ):
        """MediaPipe手部关键点检测器
        Args:
            hands_name_ls: 需要被检测到的手部名字列表
            hands_matcher: 手部匹配器,用于匹配/追踪多只手部的名字
                以下为手部检测器VisualHandDetector的参数
                static_image_mode: 使用静态检测模式,False为使用动态
                min_detect_confi: 检测手部的置信度
                min_track_confi: 跟踪手部的置信度
        """
        super().__init__(hands_name_ls, hands_matcher)
        self.hands_name_ls: list[str] = hands_name_ls
        # 初始化手部检测器
        self._detector = mp.solutions.hands.Hands(
            static_image_mode=static_image_mode,
            max_num_hands=len(hands_name_ls),
            min_detection_confidence=min_detect_confi,
            min_tracking_confidence=min_track_confi,
        )
        # 创建手部匹配器
        self.hands_matcher: HandsMatcher = hands_matcher(hands_name_ls)
        # 创建手部匹配计时器,用于记录手部识别了多少帧,初始值为10
        self.matched_counter: dict[str, int] = {n: 30 for n in hands_name_ls}

    def detect(self, image: np.ndarray, hands_dict: dict[str, OneHand]) -> list[str]:
        """使用MediaPipe检测手部关键点,返回成功检测到的手部的名称
        Args:
            image: ndarray格式的被检测图片
            hands_dict: 手部名字和手部数据实例的字典,名字为键,数据实例为值
        """
        # 用一个列表保存检测到的手部的名字
        detected_name_ls = []
        # MediaPipe检测手部关键点需要转换为RGB
        imgRGB = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        # 检测手部关键点
        results = self._detector.process(imgRGB)
        multi_hand_landmarks = results.multi_hand_landmarks
        multi_handedness = results.multi_handedness
        # 判断是否检测到手部
        if multi_hand_landmarks:
            # 计算手部相似度矩阵,行是检测到的手部编号(索引),列是需要配对的手部名称
            r_num, c_num = len(multi_hand_landmarks), len(self.hands_name_ls)
            if r_num > c_num:  # 检测的手部数量过多
                return []
            sim_matix: np.ndarray = np.empty((r_num, c_num), dtype=np.float16)
            # 计算相似度,填入对应矩阵的位置
            for detected_idx in range(r_num):
                for name_idx, name in enumerate(self.hands_name_ls):
                    one_hand = hands_dict[name]
                    landmark = multi_hand_landmarks[detected_idx].landmark
                    handedness = multi_handedness[detected_idx]
                    # 1. 手腕关键点距离相似度
                    # 计算下一帧预测位置,使用五指指尖和手腕的关键点位置来计算相似度
                    pred_ipos0 = one_hand.raw_pos[0, :2] * 2 - one_hand.last_rpos[0, :2]
                    dist_sim = np.linalg.norm(
                        (pred_ipos0 / image.shape[:2] - (landmark[0].x, landmark[0].y)),
                        ord=1,
                    )
                    # 2. 左右手标签相似度
                    # 根据左右手标签对相似度乘以比例
                    hand_side_sim = 10
                    if one_hand.hand_side == handedness.classification[0].label:
                        hand_side_sim = 0
                    elif one_hand.hand_side == "Unknown":
                        hand_side_sim = 20
                    # 3. 匹配的帧的数量
                    # 优先匹配之前匹配过的手部
                    matched_num = self.matched_counter[name]
                    # 写入对应的相似度
                    cur_similarity = 10 * dist_sim + hand_side_sim + matched_num
                    sim_matix[detected_idx, name_idx] = cur_similarity
            # 计算得到匹配手部字典
            matched_dict: dict[int, str] = self.hands_matcher.run(sim_matix)

            # 更新匹配的帧计数器
            for name in self.hands_name_ls:
                if name in matched_dict.values():
                    if self.matched_counter[name] > 0:
                        self.matched_counter[name] -= 1  # 匹配到且计数器当前大于0就减1
                else:
                    self.matched_counter[name] = 30  # 重置计数器

            # 根据匹配结果处理对应的手部数据
            img_h, img_w, _ = image.shape  # 获取图片宽高,用于处理归一化后的关键点数据
            for detected_idx, name in matched_dict.items():
                one_hand = hands_dict[name]
                # 记录该手部是左手还是右手
                one_hand.hand_side = (
                    multi_handedness[detected_idx].classification[0].label
                )
                # 记录上一帧的图片关键点位置
                one_hand.last_rpos[:, :] = one_hand.raw_pos
                # 获取并处理检测到的手部数据
                for id, landmark in enumerate(
                    multi_hand_landmarks[detected_idx].landmark
                ):
                    # 将归一化的手部关键点转化为图片中的位置
                    x = int(landmark.x * img_w)
                    y = int(landmark.y * img_h)
                    one_hand.raw_pos[id, :] = x, y, landmark.z
                # 将原始数据处理完毕后,重置所有数据计算标志
                one_hand.reset_all_flags()
                # 更新检测到的手部的名字
                detected_name_ls.append(name)
        else:
            # 重置匹配的帧计数器
            for name in self.hands_name_ls:
                self.matched_counter[name] = 30  # 重置计数器

        return detected_name_ls



# 核心功能类
from .HandInput import HandInput

# 子包核心类(常用)
from .HandUtils.Camera import Camera
from .HandInputSchemes.FingertipButtonScheme import FingertipButtonScheme
from .HandInputSchemes.ThumbJoystickScheme import ThumbJoystickScheme
from .HandInputSchemes.FingerSwipeScheme import FingerSwipeScheme

# 子包入口(进阶,保留子包结构)
from . import HandData, HandDetector, HandInputSchemes, HandUtils

# 定义版本
__version__ = "0.1.0"
# 定义公共接口
__all__ = [
    # 常用类
    "HandInput",
    "Camera",
    "FingertipButtonScheme",
    "ThumbJoystickScheme",
    "FingerSwipeScheme",
    # 子包入口
    "HandData",
    "HandDetector",
    "HandInputSchemes",
    "HandUtils",
]



from abc import ABC, abstractmethod
import numpy as np


class HandsMatcher(ABC):
    @abstractmethod
    def __init__(self, matched_ls: list[str]):
        """多手部匹配器,用于为检测到的多只手部匹配名字
        Args:
            matched_ls: 需要被匹配的手部名字列表
        """
        self.matched_ls: list[str] = matched_ls
        self.n = len(matched_ls)
        self.row_n: int = 0  # 记录当前检测到的行/数量
        self.matched_dict: dict[int, str] = dict()

    @abstractmethod
    def run(self, cost_matrix: np.ndarray) -> dict:
        """运行匹配算法,返回匹配后的字典
        Args:
            cost_matrix: 未扩展的初始的成本矩阵,列数要和被检测手部数量相同
        """
        return self.matched_dict


class HungarianMatcher(HandsMatcher):
    def __init__(self, matched_ls: list[str]):
        """用匈牙利算法为多只手部匹配名字
        Args:
            matched_ls: 需要被匹配的手部名字列表
        """
        self.matched_ls: list[str] = matched_ls
        self.n: int = len(matched_ls)
        self.row_n: int = 0  # 记录当前检测到的行/数量
        self.matched_dict: dict[int, str] = dict()

    def _extend2square(self, cost_matrix: np.ndarray) -> np.ndarray:
        """用0行填充剩下没有检测到的手,扩展成方阵,返回扩展后的成本矩阵(方阵)
        Args:
            cost_matrix: 未扩展的初始的成本矩阵,列数要和被检测手部数量相同
        """
        self.row_n, col_n = cost_matrix.shape
        if self.row_n < self.n:  # 添加没有检测到的行(虚拟行),扩展成方阵才能用该算法
            return np.vstack([cost_matrix, np.zeros((self.n - self.row_n, col_n))])
        return cost_matrix.copy()

    def _reduce_matrix(self, cur_matrix: np.ndarray) -> None:
        """矩阵归约,生成尽可能多的零元素
        Args:
            cur_matrix: 当前处理过之后的方阵
        """
        # 行归约:每行减去最小值,得到零元素
        row_mins = cur_matrix.min(axis=1, keepdims=True)
        cur_matrix -= row_mins
        # 列归约:每列减去最小值
        col_mins = cur_matrix.min(axis=0, keepdims=True)
        cur_matrix -= col_mins

    def _find_0line(self, cur_matrix: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        """用贪心策略找到最优的零覆盖线,返回标记了画线的行和列的布尔索引的元组
        Args:
            cur_matrix: 当前处理过之后的方阵
        """
        # assert cost_matrix.shape[0] == n and cost_matrix.shape[1] == n
        # 获取零元素所在位置的方阵
        zeros_matrix = cur_matrix == 0  # 得到布尔值矩阵,零元素位置为True
        # 定义划线位置,row_line记录横着画的线,col_line记录竖线
        row_line = np.zeros(self.n, dtype=np.bool)
        col_line = np.zeros(self.n, dtype=np.bool)
        while np.sum(zeros_matrix) > 0:  # 不断画线直到覆盖到全部的零元素
            # 在最多零的地方画横线
            row_line[np.argmax(np.sum(zeros_matrix, axis=1))] = True
            zeros_matrix[row_line, :] = False  # 将已经画了线的地方设为False
            if np.sum(zeros_matrix) <= 0:
                break  # 每次画完一条线都检测是否完全覆盖,横线竖线交替画
            col_line[np.argmax(np.sum(zeros_matrix, axis=0))] = True  # 画竖线
            zeros_matrix[:, col_line] = False
        return row_line, col_line

    def _adjust_matrix(
        self, cur_matrix: np.ndarray, row_line: np.ndarray, col_line: np.ndarray
    ) -> None:
        """调整矩阵,在未覆盖区域生成新的零,同时保持已有零的位置不变
        Args:
            cur_matrix: 当前处理过之后的方阵
            row_line: 行的画线标志数组
            col_line: 列的画线标志数组
        """
        # 标记覆盖区域
        covered_matrix = np.zeros_like(cur_matrix, dtype=np.bool)
        covered_matrix[row_line, :] = True  # 标记横线覆盖区域
        covered_matrix[:, col_line] = True  # 标记竖线覆盖区域
        # 获取未覆盖区域的最小值
        uncovered_min = np.min(cur_matrix[~covered_matrix])
        # 对未覆盖区域进行减法,在未覆盖区域生成新零
        cur_matrix[~covered_matrix] -= uncovered_min
        # 对覆盖线交叉的位置进行加法,防止已有零被破坏
        cur_matrix[row_line, col_line] += uncovered_min

    def _mark_zeros(self, cur_matrix: np.ndarray) -> dict[int, str]:
        """标记独立零,通过找独立零来匹配两边的节点,返回最终匹配后的字典
        Args:
            cur_matrix: 当前处理过之后的方阵
        """
        # 获取零元素所在位置的方阵
        zeros_matrix = cur_matrix == 0  # 得到布尔值矩阵,零元素位置为True
        for _ in range(self.row_n):
            # 贪心策略获取独立零
            zero_num = np.sum(zeros_matrix, axis=1)  # 统计每行零的数量
            zero_num[zero_num <= 0] = self.n  # 将已经选过的行设为最大(减少取到该行)
            zero_num[self.row_n + 1 :] = self.n  # 将没有检测到的行设为最大
            row_idx = int(np.argmin(zero_num))  # 取零数量最少的行作为本行的独立零
            col_idx = int(np.argmax(zeros_matrix[row_idx, :]))  # 获取对应的列索引
            # 清空所在行和列的0
            zeros_matrix[row_idx, :] = False
            zeros_matrix[:, col_idx] = False
            # 对应匹配,检查是否匹配到没有检测到的行
            if row_idx < self.row_n:
                self.matched_dict[row_idx] = self.matched_ls[col_idx]
        return self.matched_dict

    def run(self, cost_matrix: np.ndarray) -> dict[int, str]:
        """运行匈牙利算法,返回匹配后的字典(键为检测到手部的索引,值为对应的名字)
        Args:
            cost_matrix: 未扩展的初始的成本矩阵,列数要和被检测手部数量相同
        """
        # assert cost_matrix.ndim == 2 and cost_matrix.shape[1] == self.n
        if cost_matrix.shape[1] != self.n:
            return {}
        self.matched_dict.clear()  # 清空上次的匹配结果
        # 扩展为方阵,才能使用匈牙利算法
        matrix = self._extend2square(cost_matrix)
        self._reduce_matrix(matrix)
        row_line, col_line = self._find_0line(matrix)
        # 用覆盖线数量来判定是否完美匹配
        while (row_line.sum() + col_line.sum()) < self.n:
            self._adjust_matrix(matrix, row_line, col_line)  # 调整矩阵
            row_line, col_line = self._find_0line(matrix)  # 更新覆盖线
        # 最后通过找独立零来匹配对应的名字
        return self._mark_zeros(matrix)


if __name__ == "__main__":
    # 测试
    matched_ls = ["0", "1", "2"]
    matcher = HungarianMatcher(matched_ls)  # 创建匈牙利算法实例

    cost_matrix = np.array([[3, 1, 2], [2, 4, 7], [3, 6, 8]])
    matches = matcher.run(cost_matrix)
    print(1, matches)  # res: 0->2, 1->1, 2->0

    cost_matrix = np.arange(9).reshape(3, 3)
    matches = matcher.run(cost_matrix)
    print(2, matches)  # res: 0->0, 1->1

    cost_matrix = np.array([[3, 1, 2], [2, 4, 7]])
    matches = matcher.run(cost_matrix)
    print(3, matches)  # res: 0->1, 1->0

    cost_matrix = np.array([[2, 2, 3], [2, 2, 3], [3, 3, 4]])
    matches = matcher.run(cost_matrix)
    print(4, matches)  # res: 0->0, 1->1, 2->2 或 0->1, 1->0, 2->2

    cost_matrix = np.array([[3, -1, 2], [2, 4, 7], [-3, 6, 8]])
    matches = matcher.run(cost_matrix)
    print(5, matches)  # res: 0->1, 1->2, 2->0

    cost_matrix = np.zeros((3, 3))
    matches = matcher.run(cost_matrix)
    print(6, matches)  # res: 0->0, 1->1, 2->2等任意匹配都可以

    matched_ls = ["0", "1", "2", "3"]
    matcher = HungarianMatcher(matched_ls)  # 创建匈牙利算法实例
    cost_matrix = np.array([[5, 3, 6, 2], [4, 7, 1, 3], [2, 5, 4, 8], [9, 2, 7, 5]])
    matches = matcher.run(cost_matrix)
    print(7, matches)  # res: 0->3, 1->2, 2->0, 3->1



from typing import Literal
import numpy as np

from ..HandUtils.Filter import AdaptiveEWMAFilter
from ._global import T_NORM2WRIST, T_NORM2VEC


class OneHand:
    __slots__ = (
        "hand_side",
        "raw_pos",
        "last_rpos",
        "_norm_pos",
        "_box",
        "_normalized_flag",
        "_wrist_npos",
        "_wrist_npos_flag",
        "_fingers_angle",
        "_angle_flag",
        "_thumb_dist",
        "_dist_flag",
        "_data",
        "_data_flag",
        "_pos_data",
        "_pos_data_flag",
        "_finger_data",
        "_finger_data_flag",
        "_wrist_npos_diff",
        "last_wnpos",
        "wrist_npos_diff_filter",
        "_wrist_npos_diff_flag",
        "_angle_diff",
        "_last_angle",
        "_angle_diff_flag",
    )

    def __init__(self):
        """一只手的关键点位置数据"""
        # 用记录当前手是左手还是右手,默认为未知
        self.hand_side: Literal["left", "right", "Unknown"] = "Unknown"
        # 定义一个21x3的全0的二维数组来接收返回的21个手部关键点的xyz坐标
        # raw_pos表示这是原始传入的手部关键点的位置,初始创建一个空的数组,里面初始值是脏数据
        self.raw_pos: np.ndarray = np.zeros((21, 3), dtype=np.float32)
        self.last_rpos: np.ndarray = self.raw_pos.copy()
        # norm_pos表示是再手部box里归一化之后的坐标点坐标
        self._norm_pos: np.ndarray = np.zeros((21, 3), dtype=np.float32)
        # 手部矩形框的四个坐标点,分别为左上角坐标和右下角坐标
        self._box: np.ndarray = np.zeros(4, dtype=np.int32)
        # 标记是否已经计算好归一化和矩形框的四个坐标点
        self._normalized_flag: bool = False
        # _wrist_npos表示以手腕为原点的归一化坐标
        self._wrist_npos: np.ndarray = np.zeros((21, 3), dtype=np.float32)
        # 标记是否已经计算好以手腕为原点的归一化坐标
        self._wrist_npos_flag: bool = False
        # 定义一个5x3的二维数组用来存储5根手指的3个关节点的弯曲角度
        self._fingers_angle: np.ndarray = np.zeros((5, 3), dtype=np.float32)
        # 标记是否计算好关键点的弯曲角度
        self._angle_flag: bool = False
        # 定义一个一维数组来存储大拇指与其他关键点的距离
        self._thumb_dist: np.ndarray = np.zeros(4, dtype=np.float32)
        # 标志是否计算好4个指尖与大拇指的距离
        self._dist_flag: bool = False
        # 创建一个一维数组用于收集所有的一维手部数据
        self._data: np.ndarray = self._norm_pos.copy()
        # 标志当前数据是否为最新的数据
        self._data_flag: bool = False
        # 创建一个一维数组用于收集关键点的一维手部数据
        self._pos_data: np.ndarray = self._norm_pos.ravel()
        # 标志当前数据是否为最新的数据
        self._pos_data_flag: bool = False
        # 创建一个一维数组用于收集角度和指尖距离的一维手部数据
        self._finger_data: np.ndarray = self._fingers_angle.ravel()
        # 标志当前数据是否为最新的数据
        self._finger_data_flag: bool = False
        # 创建一个21x3的二维数组来存储每个关键点的移动差值
        self._wrist_npos_diff: np.ndarray = np.zeros(21, dtype=np.float32)
        self.last_wnpos: np.ndarray = self._wrist_npos.copy()
        # 创建滤波器对关键点两帧间的位移进行滤波
        self.wrist_npos_diff_filter: AdaptiveEWMAFilter = AdaptiveEWMAFilter(0.08, 0.5)
        # 标志当前_wrist_npos_diff变量存的是否为最新的速度
        self._wrist_npos_diff_flag: bool = False
        # 定义一个5x3的二维数组用来存储5根手指的3个关节点的弯曲角度的差值
        self._angle_diff: np.ndarray = self._fingers_angle.copy()
        self._last_angle: np.ndarray = self._fingers_angle.copy()
        # 标志当前_angle_diff变量存的是否为最新的速度
        self._angle_diff_flag: bool = False

    def reset_all_flags(self):
        """重置所有数据的更新标志为False"""
        self._normalized_flag = False
        self._wrist_npos_flag = False
        self._angle_flag = False
        self._dist_flag = False
        self._data_flag = False
        self._finger_data_flag = False
        self._wrist_npos_diff_flag = False
        self._angle_diff_flag = False

    @property
    def norm_pos(self) -> np.ndarray:
        """返回以手部矩形框左上角为原点的归一化关键点坐标"""
        if not self._normalized_flag:
            return self.normalization()
        return self._norm_pos

    def normalization(self) -> np.ndarray:
        """将传入的初始手部关键点坐标进行归一化为手部框内的相对位置"""
        # 计算得到手部矩形框的四个顶点坐标
        min_arr = self.raw_pos.min(axis=0)
        max_arr = self.raw_pos.max(axis=0)
        # 计算顶点的范围,最大减最小
        min_x, min_y, _ = min_arr
        max_x, max_y, _ = max_arr
        # 更新手部矩形框
        self._box[:] = min_x, min_y, max_x, max_y
        # 归一化计算,以手部矩形框左上角为原点,坐标值范围在0到1内
        self._norm_pos[:] = (self.raw_pos - min_arr) / (max_arr - min_arr)
        self._normalized_flag = True
        return self._norm_pos

    @property
    def wrist_npos(self) -> np.ndarray:
        """返回以手腕关键点坐标为原点的归一化关键点坐标"""
        if not self._wrist_npos_flag:
            return self.calc_wrist_npos()
        return self._wrist_npos

    def calc_wrist_npos(self) -> np.ndarray:
        """计算得到以手腕为原点的归一化坐标"""
        # 当前归一化坐标是以手部矩形框左上角为原点,坐标值范围在0到1内
        self._wrist_npos[:] = self.norm_pos
        # 以手腕关键点坐标位置为原点
        self._wrist_npos[1:, :] -= self._wrist_npos[0, :]
        self._wrist_npos_flag = True
        return self._wrist_npos

    def calc_wrist_npos2(self) -> np.ndarray:
        """计算得到以手腕为原点的归一化坐标"""
        self._wrist_npos[:] = np.dot(T_NORM2WRIST, self.norm_pos)
        self._wrist_npos_flag = True
        return self._wrist_npos

    @property
    def box(self) -> np.ndarray:
        """获取手部矩形框4个顶点坐标"""
        if not self._normalized_flag:
            self.normalization()
        return self._box

    @property
    def box_w(self) -> int:
        return self._box[3] - self._box[1]

    @property
    def box_h(self) -> int:
        return self._box[2] - self._box[0]

    @property
    def fingers_angle(self) -> np.ndarray:
        """获取手指每个关节点的弯曲角度"""
        if not self._angle_flag:
            return self.calc_5fingers_angle()
        return self._fingers_angle

    def calc_5fingers_angle(self) -> np.ndarray:
        """计算所有手指的每个关节点的弯曲角度"""
        # 使用行变换计算当前手部关键点坐标组成的向量,每个行向量代表一截手指关节向量
        fingers_vec = np.dot(T_NORM2VEC, self.norm_pos)  # 左乘行变换矩阵
        fingers_vec = fingers_vec[1:, :]  # 去掉第0行,第0行不是向量
        # 将每个手指向量都单位化,变成单位向量
        # 先计算行向量的模长
        vec_length = np.sqrt(np.sum(fingers_vec * fingers_vec, axis=1)).reshape((20, 1))
        fingers_vec = fingers_vec / vec_length  # 每个行向量元素都除以行向量模长
        # 计算每个手指的向量的夹角,这里range(0,20,4)是取每根手指最小的关键点到关键点0的向量
        for f_idx, v_idx in enumerate(range(0, 20, 4)):  # 每4截关节向量为一根手指
            finger_angle = np.acos(
                np.sum(
                    fingers_vec[v_idx : (v_idx + 3), :]
                    * fingers_vec[(v_idx + 1) : (v_idx + 4), :],
                    axis=1,
                )
            )
            # 将计算的角度结果,赋值给用于存储的变量
            self._fingers_angle[f_idx, :] = finger_angle
        self._angle_flag = True
        return self._fingers_angle

    @property
    def thumb_dist(self) -> np.ndarray:
        """获取4个指尖到大拇指的距离"""
        if not self._dist_flag:
            return self.calc_thumb_distance()
        return self._thumb_dist

    def calc_thumb_distance(self) -> np.ndarray:
        """计算大拇指指尖到其他4个手指指尖的距离"""
        thumb_tip_point = self.norm_pos[4, :]
        for i, finger_id in enumerate((8, 12, 16, 20)):
            # 用曼哈顿距离的计算量没有欧式距离大
            # 根据手指的编号获取该点的归一化的xyz坐标
            finger_point = self.norm_pos[finger_id, :]
            # 直接用l1范数来计算,也可以用np.sum(np.abs(point1 - point2))
            self._thumb_dist[i] = np.linalg.norm(thumb_tip_point - finger_point, ord=1)
        self._dist_flag = True
        return self._thumb_dist

    @property
    def data(self) -> np.ndarray:
        """获取所有已整合好的数据"""
        if not self._data_flag:
            return self.integrate_data()
        return self._data

    def integrate_data(self) -> np.ndarray:
        """整合所有手部相关数据并输出为一维数组"""
        self._data = self.norm_pos.copy()
        if self.hand_side != "right":  # 将关键点数据统一成右手数据
            self._data[:, 0] = 1 - self.norm_pos[:, 0]
        # 将数据展平然后合并
        self._data = np.ravel(self._data, order="C")
        self._data = np.concatenate(
            (self._data, np.ravel(self.fingers_angle, order="C"), self.thumb_dist)
        ).reshape(1, -1)
        self._data_flag = True
        return self._data

    @property
    def pos_data(self) -> np.ndarray:
        """获取手指角度和指尖距离数据"""
        if not self._pos_data_flag:
            self._pos_data = np.ravel(self.norm_pos, order="C").reshape(1, -1)
            self._pos_data_flag = True
        return self._pos_data

    @property
    def finger_data(self) -> np.ndarray:
        """获取手指角度和指尖距离数据"""
        if not self._finger_data_flag:
            return self.integrate_finger_data()
        return self._finger_data

    def integrate_finger_data(self) -> np.ndarray:
        """整合手指角度和指尖距离的数据并输出为一维数组"""
        self._finger_data = np.concatenate(
            (np.ravel(self.fingers_angle, order="C"), self.thumb_dist)
        ).reshape(1, -1)
        self._finger_data_flag = True
        return self._finger_data

    @property
    def wrist_npos_diff(self):
        """更新并获取两帧之间的归一化坐标差值"""
        if not self._wrist_npos_diff_flag:
            # 计算上次和本次的手部关键点距离
            self._wrist_npos_diff[:] = self.wrist_npos_diff_filter(
                np.linalg.norm(
                    self.wrist_npos[:, :2] - self.last_wnpos[:, :2], ord=1, axis=1
                )
            )
            self.last_wnpos[:, :] = self.wrist_npos  # 记录本次关键点位置
            self._wrist_npos_diff_flag = True
        return self._wrist_npos_diff

    @property
    def angle_diff(self):
        """更新并获取两帧之间手指关节角度的差值"""
        if not self._angle_diff_flag:
            # 计算上次和本次的手部关键点距离
            self._angle_diff = np.abs(self.fingers_angle - self._last_angle)
            self._last_angle[:, :] = self.fingers_angle  # 记录本次关键点位置
            self._angle_diff_flag = True
        return self._angle_diff



import numpy as np


class EWMAFilter:
    __slots__ = "smooth_factor", "x"

    # 基本的指数加权移动平均 (Basic EWMA)
    def __init__(self, smooth_factor: float = 0.5):
        # 平滑因子越接近0平滑效果越强.但响应速度更慢;越接近1响应速度越快,但平滑效果较弱
        self.smooth_factor: float = smooth_factor  # 设定平滑因子
        # 这里的平滑因子可以设置成一个向量,来更新整个矩阵,所以没有指定类型
        self.x = None  # 未设定初始值

    def update(self, new_x, new_smooth: None | float = None):
        if self.x is None:
            self.x = new_x  # 设定初始值
        else:  # 根据公式进行平滑
            if new_smooth is not None:  # 使用新的平滑因子
                self.smooth_factor = new_smooth
            self.x = self.smooth_factor * new_x + (1 - self.smooth_factor) * self.x
        return self.x

    def __call__(self, new_x, new_smooth: None | float = None):
        return self.update(new_x, new_smooth)

    def decorator(self, func):
        """装饰器,将装饰函数的返回结果进行过滤"""

        def wrapper(*args, **kwargs):
            return self.update(func(*args, **kwargs))

        return wrapper


class AdaptiveEWMAFilter:
    __slots__ = "smooth_factor", "learning_rate", "x", "last_error"

    # 自适应指数平滑 (Adaptive Exponential Smoothing)
    def __init__(self, smooth_factor: float = 0.5, learning_rate: float = 0.1):
        self.smooth_factor: float = smooth_factor  # 设定平滑因子的初始值
        # 自适应指数平滑的核心思想是根据当前误差和前一时刻的误差来调整平滑因子
        self.learning_rate: float = learning_rate  # 设定学习率
        # 学习率越大,调整得越快;学习率越小,调整得越慢
        self.x = None  # 未设定初始值
        self.last_error = 0  # 记录上一次的预测误差

    def update(self, new_x):
        if self.x is None:
            self.x = new_x  # 设定初始值
        else:
            # 计算当前误差,当前误差=当前观测值-上一次的预测值(就是self.x)
            cur_error = np.abs(new_x - self.x)
            # 使用学习率和误差来计算平滑因子
            self.smooth_factor += self.learning_rate * (cur_error - self.last_error)
            # 确保更新后的平滑因子的值在0到1间
            self.smooth_factor = np.clip(self.smooth_factor, 0, 1)
            # 更新过滤后的值
            self.x = self.smooth_factor * new_x + (1 - self.smooth_factor) * self.x
            # 更新上一次的误差
            self.last_error = cur_error
        return self.x

    def __call__(self, new_x):
        return self.update(new_x)

    def decorator(self, func):
        """装饰器,将装饰函数的返回结果进行过滤"""

        def wrapper(*args, **kwargs):
            return self.update(func(*args, **kwargs))

        return wrapper



from abc import ABC, abstractmethod

import numpy as np

from .HandsMatchers import HandsMatcher
from ..HandData.OneHand import OneHand


class VisualHandDetector(ABC):
    __slots__ = "hands_name_ls", "_detector", "hands_matcher"

    @abstractmethod
    def __init__(self, hands_name_ls: list[str], hands_matcher: type[HandsMatcher]):
        """视觉手部关键点检测器
        Args:
            hands_name_ls: 需要被检测到的手部名字列表
            hands_matcher: 手部匹配器,用于匹配/追踪多只手部的名字
        """
        # 创建手部关键点检测器
        self._detector = None
        self.hands_name_ls: list[str] = hands_name_ls
        self.hands_matcher: HandsMatcher = hands_matcher(hands_name_ls)

    @abstractmethod
    def detect(self, image: np.ndarray, hands_dict: dict[str, OneHand]) -> list[str]:
        """使用视觉检测手部关键点,返回成功检测到的手部的名称
        Args:
            image: ndarray格式的被检测图片
            hands_dict: 手部名字和手部数据实例的字典,名字为键,数据实例为值
        """
        detected_name_ls = []
        # 最后返回检测到的手部的名字
        return detected_name_ls



from time import time

import numpy as np

from .HandInputScheme import HandInputScheme
from ..HandInput import HandInput


class FingertipButtonScheme(HandInputScheme):
    __slots__ = (
        "hand_input",
        "hand_name",
        "finger_touch_idx",
        "short_press_time",
        "long_press_time",
        "_start_time",
        "_start_point",
        "_short_press_flag",
        "_press_status",
    )

    def __init__(
        self,
        hand_input: HandInput,
        hand_name: str,
        finger_touch_idx: int = 0,
        short_press_time: float = 0.1,
        long_press_time: float = 0.3,
    ) -> None:
        """指尖按钮操控方案
        Args:
            hand_input: 手部输入类的实例
            hand_name: 手部名字,指定为哪只手制定手部操控方案
            finger_touch_idx: 指尖触碰状态模型的索引,指定使用哪个指尖作为指尖按钮,默认为0(即食指触碰)
            short_press_time: 短按时间阈值,即短按的触发时间,默认为0.1秒
            long_press_time: 长按时间阈值,即长按的触发时间,默认为0.3秒
        """
        hand_input.hands_dict[hand_name]  # 获取该手部,没有该手部名字则报错
        self.hand_input: HandInput = hand_input
        self.hand_name: str = hand_name
        # 指尖按钮操控方案相关参数
        self.finger_touch_idx: int = finger_touch_idx
        self.short_press_time: float = short_press_time
        self.long_press_time: float = long_press_time
        # 开始按下的时间戳
        self._start_time: None | float = None
        # 开始按下的位置
        self._start_point: None | np.ndarray = None
        # 用于记录是否短按的标志
        self._short_press_flag: bool = False
        # 按键状态标志,None表示没有按,False表示长按,True表示短按
        self._press_status: None | bool = None

    def update(self) -> None | bool:
        """实时更新指尖按钮计时判断"""
        # 获取并判断该手部数据是否有被检测到
        hand = self.hand_input.hand(self.hand_name)
        if hand is None:  # 没有检测到对应的手部
            self._press_status = None  # 将按钮状态重置为没有按下
            return None
        # 检测手指指尖是否触碰
        if hand.gestrue.get_fg_touch(self.finger_touch_idx):
            if self._start_time is None:  # 记录开始触碰的时间戳
                self._start_time = time()
            else:  # 记录按下的瞬间大拇指的相对于手腕的归一化坐标
                if self._start_point is None:
                    self._start_point = hand.base.wrist_npos(4)
                # 通过按下的时间来判断是长按还是短按
                press_time = time() - self._start_time
                if press_time > self.long_press_time:  # 判断长按
                    self._short_press_flag = False  # 超过短按时间,不是短按
                    self._press_status = False  # 返回长按
                    return self._press_status
                elif press_time > self.short_press_time:  # 判断短按
                    self._short_press_flag = True  # 记录短按标志
        else:  # 重置开始按下的时间戳和记录的点
            self._start_time = None
            self._start_point = None
            if self._short_press_flag:  # 判断是否确定短按
                self._short_press_flag = False  # 重置短按标志
                self._press_status = True  # 短按
                return self._press_status
            else:
                self._press_status = None  # 没有按下
                return self._press_status

    @property
    def is_activate(self) -> bool:
        """指尖按钮是否有激活,True为激活,False为没有激活"""
        return self._press_status is not None

    @property
    def is_long_press(self) -> bool:
        """是否为长按,True为长按,False为短按或没有按下"""
        return self._press_status is False

    @property
    def is_short_press(self) -> bool:
        """是否为短按,True为短按,False为长按或没有按下"""
        return self._press_status is True

    @property
    def is_press(self) -> bool:
        """是否有按下,True为按下,False为没有按下"""
        return self._press_status is not None

    @property
    def start_point(self) -> np.ndarray | None:
        """返回最开始按下按钮的位置"""
        return self._start_point



from time import time
from typing import Generator

import numpy as np
import cv2

from .Filter import EWMAFilter


class Camera:
    __slots__ = (
        "capture",
        "_prev_time",
        "_delta_time",
        "delta_time_is_updated",
        "_fps",
        "fps_is_updated",
        "fps_filter",
    )

    def __init__(
        self,
        camera_idx: int = 0,
        max_fps: int = 30,
        cam_w: int = 640,
        cam_h: int = 480,
        smooth_factor: float = 0.08,
    ) -> None:
        self.capture = cv2.VideoCapture(camera_idx)  # 获取摄像头
        # 设置摄像头的打开参数,如果参数有修改才进行设置,否则跳过设置阶段
        # 设置参数操作很会大大延长打开摄像头的时间
        if max_fps != self.capture.get(cv2.CAP_PROP_FPS) and max_fps > 0:
            self.capture.set(cv2.CAP_PROP_FPS, max_fps)
        if cam_w != self.capture.get(cv2.CAP_PROP_FRAME_WIDTH) and cam_w > 0:
            self.capture.set(cv2.CAP_PROP_FRAME_WIDTH, cam_w)
        if cam_h != self.capture.get(cv2.CAP_PROP_FRAME_HEIGHT) and cam_h > 0:
            self.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, cam_h)
        # 记录当前一帧结束的时间
        self._prev_time: float = time()
        # 记录两帧时间差
        self._delta_time: float = 0
        self.delta_time_is_updated: bool = False
        # 记录当前帧率
        self._fps: int = 0
        self.fps_is_updated: bool = False
        # 创建帧率的滤波器,让帧率显示更加稳定
        self.fps_filter: EWMAFilter = EWMAFilter(smooth_factor)

    @property
    def frame_width(self):
        """获取帧的宽"""
        return self.capture.get(cv2.CAP_PROP_FRAME_WIDTH)

    @property
    def frame_height(self):
        """获取帧的高"""
        return self.capture.get(cv2.CAP_PROP_FRAME_HEIGHT)

    @property
    def frame_size(self) -> tuple[int, int]:
        """获取当前摄像头打开的宽高"""
        cam_w = int(self.capture.get(cv2.CAP_PROP_FRAME_WIDTH))
        cam_h = int(self.capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
        return cam_w, cam_h

    def set_camera_index(self, camera_idx: int) -> None:
        """重新设置摄像头索引"""
        self.capture = cv2.VideoCapture(camera_idx)

    def read(
        self,
        hflip: bool = True,
        key_ackii: int = -1,
        delay_ms: int = 1,
    ) -> Generator[np.ndarray, None, None]:
        """使用生成器获取摄像头的当前帧"""
        while True:
            # 读取摄像头的图片
            success, image = self.capture.read()
            if not success:  # 是否读取帧成功
                break  # 读取失败则结束循环
            self.fps_is_updated = False  # 标记帧率未更新
            self.delta_time_is_updated = False  # 标记两帧时间差未更新
            # 如果hflip为True则每次返回翻转后的图片
            yield cv2.flip(image, 1) if hflip else image
            # 检查按键输入来结束循环,waitKey返回-1表示没有按键输入
            if cv2.waitKey(delay_ms) != key_ackii:
                break  # 结束循环

    @property
    def fps(self) -> int:
        """计算当前帧率"""
        if self.fps_is_updated:
            return self._fps
        self._fps = int(self.fps_filter(int(1 / self.delta_time)))
        self.fps_is_updated = True  #  标记fps已更新
        return self._fps

    @property
    def prev_time(self):
        return self._prev_time

    @property
    def delta_time(self):
        """记录两帧之间的间隔时间"""
        if self.delta_time_is_updated:
            return self._delta_time
        cur_time = time()  # 获取当前时间
        self._delta_time = cur_time - self.prev_time  # 计算两帧之间的用时
        self._prev_time = cur_time  # 更新最后时间
        self.delta_time_is_updated = True  # 标记两帧时间差已更新
        return self._delta_time

    def draw_fps(
        self,
        image: np.ndarray,
        location: tuple[int, int] = (10, 70),
        color: tuple[int, int, int] = (255, 0, 0),
    ):
        """将帧率绘制到图片上"""
        # 将帧率写到图片上
        cv2.putText(image, str(self.fps), location, cv2.FONT_HERSHEY_PLAIN, 3, color, 3)

    def __del__(self):
        """销毁当前实例后,释放摄像头"""
        self.capture.release()

    def gaussian_box_mask(
        self,
        img: np.ndarray,
        box_ls: list[tuple[int, int, int, int]],
        kernel_size: int = 51,
        padding: int = 10,
    ):
        """对图片中box矩形框外的图片进行模糊效果"""
        # 对图片进行大核的高斯滤波来实现模糊效果
        blurred = cv2.GaussianBlur(img, (kernel_size, kernel_size), 0)
        mask = np.zeros_like(img)  # 创建初始的掩码
        for box in box_ls:  # 遍历传入的box_ls来创建掩码
            x0, y0, x1, y1 = box
            mask[y0 - padding : y1 + padding, x0 - padding : x1 + padding, :] = 255
        # 应用掩码,只有掩码为黑色的区域应用模糊效果
        img[np.where(mask == (0, 0, 0))] = blurred[np.where(mask == (0, 0, 0))]
        return img

    def median_box_mask(
        self,
        img: np.ndarray,
        box_ls: list[tuple[int, int, int, int]],
        kernel_size: int = 51,
        padding: int = 10,
    ):
        """对图片中box矩形框外的图片进行模糊效果"""
        # 对图片进行大核的中值滤波来实现模糊效果
        blurred = cv2.medianBlur(img, kernel_size)
        mask = np.zeros_like(img)  # 创建初始的掩码
        for box in box_ls:  # 遍历传入的box_ls来创建掩码
            x0, y0, x1, y1 = box
            mask[y0 - padding : y1 + padding, x0 - padding : x1 + padding, :] = 255
        # 应用掩码,只有掩码为黑色的区域应用模糊效果
        img[np.where(mask == (0, 0, 0))] = blurred[np.where(mask == (0, 0, 0))]
        return img

    def color_box_mask(
        self,
        img: np.ndarray,
        box_ls: list[tuple[int, int, int, int]],
        mask_color: tuple[int, int, int] = (0, 0, 0),
        padding: int = 10,
    ):
        """对图片中box矩形框外的图片进行遮挡效果"""
        # 创建纯色的遮挡图
        shadowed = np.ones_like(img)
        shadowed[:, :, 0] *= mask_color[2]  # 为背景图上色
        shadowed[:, :, 1] *= mask_color[1]
        shadowed[:, :, 2] *= mask_color[0]
        mask = np.zeros_like(img)  # 创建初始的掩码
        for box in box_ls:  # 遍历传入的box_ls来创建掩码
            x0, y0, x1, y1 = box
            mask[y0 - padding : y1 + padding, x0 - padding : x1 + padding, :] = 255
        # 应用掩码,只有掩码为黑色的区域应用遮挡效果
        img[np.where(mask == (0, 0, 0))] = shadowed[np.where(mask == (0, 0, 0))]
        return img


if __name__ == "__main__":
    t1 = time()
    cam = Camera()
    print(cam.frame_size)
    print(time() - t1)
    for img in cam.read():
        cam.draw_fps(img)
        cv2.imshow("test", img)



from time import time
import numpy as np
from .FingerModels import out_model, touch_model, close_model
from .OneHand import OneHand


class Gestrue:
    __slots__ = (
        "one_hand",
        "interval_time",
        "fom_output",
        "fom_interval_timer",
        "ftm_output",
        "ftm_interval_timer",
        "fcm_output",
        "fcm_interval_timer",
    )

    def __init__(
        self,
        one_hand: OneHand,
        interval_time: float = 0.3,
    ) -> None:
        """调用手势识别模型API
        Args:
            one_hand: 输入基本的手部数据实例
            interval_time: 调用模型的间隔时间
        """
        self.one_hand: OneHand = one_hand
        # 间隔调用参数
        self.interval_time: float = interval_time
        # 检测手指是否伸出
        self.fom_output: np.ndarray = np.zeros(5, dtype=np.bool_)
        self.fom_interval_timer: list[float] = [0 for _ in range(5)]
        # 检测指尖是否触碰
        self.ftm_output: np.ndarray = np.zeros(4, dtype=np.bool_)
        self.ftm_interval_timer: list[float] = [0 for _ in range(4)]
        # 检测手指是否合拢
        self.fcm_output: np.ndarray = np.zeros(4, dtype=np.bool_)
        self.fcm_interval_timer: list[float] = [0 for _ in range(4)]
        # self.fcm_interval_timer: np.ndarray = np.zeros(4, dtype=np.float32)

    @property
    def fg_all_out(self) -> np.ndarray:
        """检测是否所有手指都伸出"""
        for i in range(self.fom_output.shape[0]):
            self.get_fg_out(i)
        return self.fom_output

    def get_fg_out(self, idx: int) -> np.ndarray:
        """检测指定索引的手指是否伸出
        Args:
            idx: 0到4分别为大拇指,食指,中指,无名指,小拇指
        """
        if idx < 0 or idx > 4:
            raise ValueError(f"There is no finger with index {idx}")
        # 计算当前间隔时间
        cur_interval = time() - self.fom_interval_timer[idx]
        # 长间隔时间:模型输出结果没有变化
        if cur_interval > self.interval_time:
            self.fom_interval_timer[idx] = time()  # 更新计时器
            # 调用模型进行推理
            self.fom_output[idx] = out_model.run_by_idx(idx, self.one_hand.pos_data)
        return self.fom_output[idx]

    @property
    def thumb_out(self) -> np.ndarray:
        """检测大拇指是否伸出"""
        return self.get_fg_out(0)

    @property
    def index_fg_out(self) -> np.ndarray:
        """检测食指是否伸出"""
        return self.get_fg_out(1)

    @property
    def middle_fg_out(self) -> np.ndarray:
        """检测中指是否伸出"""
        return self.get_fg_out(2)

    @property
    def ring_fg_out(self) -> np.ndarray:
        """检测无名指是否伸出"""
        return self.get_fg_out(3)

    @property
    def pinky_out(self) -> np.ndarray:
        """检测小拇指是否伸出"""
        return self.get_fg_out(4)

    @property
    def fg_all_touch(self) -> np.ndarray:
        """检测是否所有手指指尖都触碰"""
        for i in range(self.ftm_output.shape[0]):
            self.get_fg_touch(i)
        return self.ftm_output

    def get_fg_touch(self, idx: int) -> np.ndarray:
        """检测指定索引的手指指尖是否触碰大拇指指尖
        Args:
            idx: 0到3分别为大拇指到食指/中指/无名指/小拇指指尖的索引
        """
        if idx < 0 or idx > 3:
            raise ValueError(f"There is no finger with index {idx}")
        # 计算当前间隔时间
        cur_interval = time() - self.ftm_interval_timer[idx]
        # 长间隔时间:模型输出结果没有变化
        if cur_interval > self.interval_time:
            self.ftm_interval_timer[idx] = time()  # 更新计时器
            # 调用模型进行推理
            self.ftm_output[idx] = touch_model.run_by_idx(
                idx, self.one_hand.finger_data
            )
        return self.ftm_output[idx]

    @property
    def index_fg_touch(self) -> np.ndarray:
        """检测食指指尖是否触碰大拇指指尖"""
        return self.get_fg_touch(0)

    @property
    def middle_fg_touch(self) -> np.ndarray:
        """检测中指指尖是否触碰大拇指指尖"""
        return self.get_fg_touch(1)

    @property
    def ring_fg_touch(self) -> np.ndarray:
        """检测无名指指尖是否触碰大拇指指尖"""
        return self.get_fg_touch(2)

    @property
    def pinky_touch(self) -> np.ndarray:
        """检测小拇指指尖是否触碰大拇指指尖"""
        return self.get_fg_touch(3)

    @property
    def fg_all_close(self) -> np.ndarray:
        """检测是否所有手指都合拢"""
        for i in range(self.fcm_output.shape[0]):
            self.get_fg_out(i)
        return self.fcm_output

    def get_fg_close(self, idx: int) -> np.ndarray:
        """检测指定索引的手指之间是否并拢
        Args:
            idx: 0到3分别为大拇指和食指并拢,食指和中指并拢...的索引
        """
        if idx < 0 or idx > 3:
            raise ValueError(f"There is no finger with index {idx}")
        # 计算当前间隔时间
        cur_interval = time() - self.fcm_interval_timer[idx]
        # 长间隔时间:模型输出结果没有变化
        if cur_interval > self.interval_time:
            self.fcm_interval_timer[idx] = time()  # 更新计时器
            # 调用模型进行推理
            self.fcm_output[idx] = close_model.run_by_idx(idx, self.one_hand.pos_data)
        return self.fcm_output[idx]

    @property
    def tb_if_close(self) -> np.ndarray:
        """检测大拇指和食指是否并拢"""
        return self.get_fg_close(0)

    @property
    def if_mf_close(self) -> np.ndarray:
        """检测食指和中指是否并拢"""
        return self.get_fg_close(1)

    @property
    def mf_rf_close(self) -> np.ndarray:
        """检测中指和无名指是否并拢"""
        return self.get_fg_close(2)

    @property
    def rf_pk_close(self) -> np.ndarray:
        """检测无名指和小拇指是否并拢"""
        return self.get_fg_close(3)



"""
封装onnx模型
目前只支持使用onnx模型
"""

from os import path
from collections import namedtuple
from abc import ABC, abstractmethod

import onnxruntime
from numpy import ndarray


def load_onnx(model_file_name):
    """加载onnx模型
    Args:
        model_file_name: onnx模型文件路径
    """
    model_path = path.join(path.dirname(__file__), "models", model_file_name)
    model_path = path.abspath(model_path)
    return onnxruntime.InferenceSession(model_path)


ModelData = namedtuple("ModelData", "model mask")


class FingerModel(ABC):
    @abstractmethod
    def run_by_idx(self, idx: int, X: ndarray) -> ndarray:
        """通过输出索引来指定需要运行的手指状态模型
        Args:
            idx: 模型输出的索引
            X: 模型输入
        """
        pass

    @abstractmethod
    def run_all(self, X: ndarray) -> list[ndarray]:
        """运行该手部的所有手指状态的模型
        Args:
            X: 模型输入
        """
        pass


class FingerOutModel(FingerModel):
    __slots__ = "output"

    def __init__(self) -> None:
        """手指伸出状态模型"""
        # 定义手指伸出状态模型的输出
        FingerOutOutput = namedtuple(
            "FingerOutOutput", "thumb index_fg middle_fg ring_fg pinky"
        )
        # 加载模型和指定模型输入
        self.output = FingerOutOutput(
            thumb=ModelData(
                model=load_onnx("tb_out_model.onnx"),
                mask="pos_data",
            ),
            index_fg=ModelData(
                model=load_onnx("if_out_model.onnx"),
                mask="pos_data",
            ),
            middle_fg=ModelData(
                model=load_onnx("mf_out_model.onnx"),
                mask="pos_data",
            ),
            ring_fg=ModelData(
                model=load_onnx("rf_out_model.onnx"),
                mask="pos_data",
            ),
            pinky=ModelData(
                model=load_onnx("pk_out_model.onnx"),
                mask="pos_data",
            ),
        )

    def run_by_idx(self, idx: int, X: ndarray) -> ndarray:
        """通过输出索引来指定需要运行的手指伸出状态模型
        Args:
            idx: 0到4分别为大拇指,食指,中指,无名指,小拇指
            X: 模型输入,所有手部关键点的位置数据
        """
        # X = X[self.output[idx].mask]
        return self.output[idx].model.run(["label"], {"f32X": X})[0]

    def run_all(self, X: ndarray) -> list[ndarray]:
        """运行该手部的所有手指伸出状态的模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        # return [fg.model.run(["label"], {"f32X": X[fg.mask]})[0] for fg in self.output]
        return [fg.model.run(["label"], {"f32X": X})[0] for fg in self.output]

    def thumb(self, X: ndarray) -> ndarray:
        """运行大拇指伸出状态模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        return self.output.thumb.model.run(["label"], {"f32X": X})[0]

    def index_fg(self, X: ndarray) -> ndarray:
        """运行食指伸出状态模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        return self.output.index_fg.model.run(["label"], {"f32X": X})[0]

    def middle_fg(self, X: ndarray) -> ndarray:
        """运行中指伸出状态模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        return self.output.middle_fg.model.run(["label"], {"f32X": X})[0]

    def ring_fg(self, X: ndarray) -> ndarray:
        """运行无名指伸出状态模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        return self.output.ring_fg.model.run(["label"], {"f32X": X})[0]

    def pinky(self, X: ndarray) -> ndarray:
        """运行小拇指伸出状态模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        return self.output.pinky.model.run(["label"], {"f32X": X})[0]


class FingerTouchModel(FingerModel):
    __slots__ = "output"

    def __init__(self) -> None:
        # 定义指尖触碰状态的输出
        FingerTouchOutput = namedtuple(
            "FingerTouchOutput", "index_fg middle_fg ring_fg pinky"
        )
        # 加载模型和定义模型输入
        self.output = FingerTouchOutput(
            index_fg=ModelData(
                model=load_onnx("if_touch_model.onnx"),
                mask="finger_data",
            ),
            middle_fg=ModelData(
                model=load_onnx("mf_touch_model.onnx"),
                mask="finger_data",
            ),
            ring_fg=ModelData(
                model=load_onnx("rf_touch_model.onnx"),
                mask="finger_data",
            ),
            pinky=ModelData(
                model=load_onnx("pk_touch_model.onnx"),
                mask="finger_data",
            ),
        )

    def run_by_idx(self, idx: int, X: ndarray) -> ndarray:
        """通过输出索引来指定需要运行的指尖触碰状态模型
        Args:
            idx: 0到3分别为大拇指到食指/中指/无名指/小拇指指尖的索引
            X: 模型输入,包括所有手指弯曲和指尖距离数据
        """
        return self.output[idx].model.run(["label"], {"f32X": X})[0]

    def run_all(self, X: ndarray) -> list[ndarray]:
        """运行该手部的所有指尖触碰状态的模型
        Args:
            X: 模型输入,包括所有手指弯曲和指尖距离数据
        """
        return [fg.model.run(["label"], {"f32X": X})[0] for fg in self.output]

    def index_fg(self, X: ndarray) -> ndarray:
        """运行大拇指到食指的指尖触碰状态模型
        Args:
            X: 模型输入,包括所有手指弯曲和指尖距离数据
        """
        return self.output.index_fg.model.run(["label"], {"f32X": X})[0]

    def middle_fg(self, X: ndarray) -> ndarray:
        """运行大拇指到中指的指尖触碰状态模型
        Args:
            X: 模型输入,包括所有手指弯曲和指尖距离数据
        """
        return self.output.middle_fg.model.run(["label"], {"f32X": X})[0]

    def ring_fg(self, X: ndarray) -> ndarray:
        """运行大拇指到无名指的指尖触碰状态模型
        Args:
            X: 模型输入,包括所有手指弯曲和指尖距离数据
        """
        return self.output.ring_fg.model.run(["label"], {"f32X": X})[0]

    def pinky(self, X: ndarray) -> ndarray:
        """运行大拇指到小拇指的指尖触碰状态模型
        Args:
            X: 模型输入,包括所有手指弯曲和指尖距离数据
        """
        return self.output.pinky.model.run(["label"], {"f32X": X})[0]


class FingerCloseModel(FingerModel):
    __slots__ = "output"

    def __init__(self) -> None:
        # 定义手指并拢状态模型的输出
        FingerCloseOutput = namedtuple("FingerCloseOutput", "tb_if if_mf mf_rf rf_pk")
        # 加载模型和指定模型输入
        self.output = FingerCloseOutput(
            tb_if=ModelData(
                model=load_onnx("ti_close_model.onnx"),
                mask="pos_data",
            ),
            if_mf=ModelData(
                model=load_onnx("im_close_model.onnx"),
                mask="pos_data",
            ),
            mf_rf=ModelData(
                model=load_onnx("mr_close_model.onnx"),
                mask="pos_data",
            ),
            rf_pk=ModelData(
                model=load_onnx("rp_close_model.onnx"),
                mask="pos_data",
            ),
        )

    def run_by_idx(self, idx: int, X: ndarray) -> ndarray:
        """通过输出索引来指定需要运行的手指并拢状态模型
        Args:
            idx: 0到3分别为大拇指和食指并拢,食指和中指并拢...的索引
            X: 模型输入,所有手部关键点的位置数据
        """
        return self.output[idx].model.run(["label"], {"f32X": X})[0]

    def run_all(self, X: ndarray) -> list[ndarray]:
        """运行该手部的所有手指并拢状态的模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        return [fg.model.run(["label"], {"f32X": X})[0] for fg in self.output]

    def tb_if(self, X: ndarray) -> ndarray:
        """运行大拇指和食指并拢状态模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        return self.output.tb_if.model.run(["label"], {"f32X": X})[0]

    def if_mf(self, X: ndarray) -> ndarray:
        """运行食指和中指并拢状态模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        return self.output.if_mf.model.run(["label"], {"f32X": X})[0]

    def mf_rf(self, X: ndarray) -> ndarray:
        """运行中指和无名指并拢状态模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        return self.output.mf_rf.model.run(["label"], {"f32X": X})[0]

    def rf_pk(self, X: ndarray) -> ndarray:
        """运行无名指和小拇指并拢状态模型
        Args:
            X: 模型输入,所有手部关键点的位置数据
        """
        return self.output.rf_pk.model.run(["label"], {"f32X": X})[0]


# 直接定义全局变量来实现单例模式
out_model = FingerOutModel()
touch_model = FingerTouchModel()
close_model = FingerCloseModel()



import numpy as np

from .HandInputScheme import HandInputScheme
from .FingertipButtonScheme import FingertipButtonScheme
from ..HandInput import HandInput


class ThumbJoystickScheme(HandInputScheme):
    __slots__ = "hand_input", "hand_name", "finger_btn", "_fixed_point", "_activate"

    def __init__(
        self, hand_input: HandInput, hand_name: str, finger_btn: FingertipButtonScheme
    ) -> None:
        """拇指摇杆操控方案
        Args:
            hand_input: 手部输入类的实例
            hand_name: 手部名字,指定为哪只手制定手部操控方案
            finger_btn: 一个用于长按设置定点的指尖按钮实例
        """
        hand_input.hands_dict[hand_name]  # 获取该手部,没有该手部名字则报错
        self.hand_input: HandInput = hand_input
        self.hand_name: str = hand_name
        # 需要使用长按来设置定点,所以要传入一个指尖按钮实例
        self.finger_btn: FingertipButtonScheme = finger_btn
        # 创建一个变量来存储定点位置,初始值为None,表示还没有定点
        self._fixed_point: np.ndarray = np.zeros((1, 3))
        # 定义一个变量用于区分,摇杆是否激活,默认为False未激活
        self._activate: bool = False

    def update(self) -> None:
        """实时更新定点的位置,直到长按才定下来"""
        # 获取并判断是否有检测到手部
        base = self.hand_input.base(self.hand_name)
        if base is None:  # 没有检测到改手部,则不激活摇杆
            self._activate = False
            return
        # 判断是否激活摇杆
        if self.finger_btn.is_long_press:  # 持续长按,则继续激活摇杆
            self._activate = True
        else:  # 长按结束,则关闭摇杆,即不激活
            self._activate = False
        # 摇杆未激活,则定点随拇指指尖移动
        if not self._activate:
            self._fixed_point = base.wrist_npos(4)  # 定点随拇指移动

    @property
    def is_activate(self) -> bool:
        """摇杆是否激活或是否启用,激活为True"""
        return self._activate

    @property
    def fixed_point(self) -> np.ndarray | None:
        """返回设置的定点位置,没长按时定点位置随拇指移动,若未设置定点则返回None"""
        return self._fixed_point if self._activate else None

    @property
    def end_point(self) -> np.ndarray | None:
        """返回摇杆的终点坐标,是相对于手腕的归一化后坐标值,没有设定点则返回None"""
        base = self.hand_input.base(self.hand_name)
        if base is not None and self._activate:  # 手部存在且摇杆已激活
            return base.wrist_npos(4)  # 终点坐标就是当前手部的拇指坐标

    @property
    def vector(self) -> np.ndarray | None:
        """返回拇指摇杆的向量(未归一化),即当前拇指的位置和定点之差"""
        end_p = self.end_point
        return (end_p - self._fixed_point) if end_p is not None else None

    @property
    def norm_vec(self) -> np.ndarray | None:
        """返回拇指摇杆的方向向量,即归一化的向量"""
        vec = self.vector
        if vec is None:
            return None
        vec_len = np.linalg.norm(vec)
        vec_len = vec_len if vec_len > 0 else 1
        return vec / vec_len



import numpy as np
import cv2

from ..HandData.OneHand import OneHand


class HandImage:
    __slots__ = "data", "_size", "_padding"

    def __init__(self, data: np.ndarray, padx: int = 0, pady: int = 0) -> None:
        """用于方便绘制归一化后的手部数据的背景图片类
        Args:
            data: 图片的ndarray
            padx: 手部数据在x轴上绘制的最大边界
            pady: 手部数据在y轴上绘制的最大边界
        """
        self.data: np.ndarray = data
        self._size: np.ndarray = np.array((self.width, self.height))
        self._padding: np.ndarray = np.array((padx, pady))

    @property
    def width(self) -> int:
        """获取图片的宽"""
        return self.data.shape[1]

    @property
    def height(self) -> int:
        """获取图片的高"""
        return self.data.shape[0]

    @property
    def size(self) -> np.ndarray:
        """获取图片的宽和高,返回ndarray格式的数据"""
        return self._size

    @property
    def padding(self) -> np.ndarray:
        """获取x轴和y轴的最大边界,返回ndarray格式的数据"""
        return self._padding

    def calc_norm2img_pos(self, norm_pos: np.ndarray) -> np.ndarray:
        """计算归一化后的手部数据在图片上的位置坐标
        Args:
            norm_pos: 归一化后的手部数据
        """
        # 保证输入的手部归一化数据是二维平面的坐标数据
        if len(norm_pos.shape) == 1 and norm_pos.shape[0] > 2:
            norm_pos = norm_pos[:2]
        # 计算在图片上的位置坐标
        img_pos = (norm_pos * (self._size - (2 * self._padding))) + self._padding
        return img_pos.astype(np.int32)


class HandBackground(HandImage):
    def __init__(
        self,
        width: int,
        height: int,
        color: tuple[int, int, int] = (0, 0, 0),
        padx: int = 0,
        pady: int = 0,
    ) -> None:
        """纯色背景图片
        Args:
            width: 背景宽度
            height: 背景高度
            color: 背景颜色
            padx: 手部数据在x轴上绘制的最大边界
            pady: 手部数据在y轴上绘制的最大边界
        """
        # 创建纯色背景图
        data = np.empty((height, width, 3), dtype=np.uint8)
        data[:, :, :] = color
        super().__init__(data, padx, pady)


class HandDrawing:
    __slots__ = (
        "one_hand",
        "point_radius",
        "point_color",
        "point_thickness",
        "line_color",
        "line_thickness",
        "raw_img",
    )

    def __init__(
        self,
        one_hand: OneHand,
    ):
        self.one_hand = one_hand
        # 绘制手部关键点参数
        self.point_radius: int = 4
        self.point_color: tuple[int, int, int] = (255, 255, 255)
        self.point_thickness: int = 3
        # 绘制手部关键点连线参数
        self.line_color: tuple[int, int, int] = (255, 255, 255)
        self.line_thickness: int = 3
        # 记录帧图像
        self.raw_img: None | np.ndarray = None

    def update(self, raw_img: np.ndarray):
        """更新帧图像"""
        self.raw_img = raw_img

    def reset(self):
        """重置帧图像,将帧图像设为None"""
        self.raw_img = None

    def set_param(
        self,
        *,
        point_radius: int = 4,
        point_color: tuple[int, int, int] = (255, 255, 255),
        point_thickness: int = 3,
        line_color: tuple[int, int, int] = (255, 255, 255),
        line_thickness: int = 3,
    ):
        """设置手部绘制相关参数
        Args:
            point_radius: 关键点半径
            point_color: 关键点颜色
            point_thickness: 关键点粗细
            line_color: 关键点连接线的线条颜色
            line_thickness: 关键点连接线的粗细
        """
        # 绘制手部关键点参数
        self.point_radius: int = point_radius
        self.point_color: tuple[int, int, int] = point_color
        self.point_thickness: int = point_thickness
        # 绘制手部关键点连线参数
        self.line_color: tuple[int, int, int] = line_color
        self.line_thickness: int = line_thickness

    def set_point(
        self,
        *,
        radius: int = 4,
        color: tuple[int, int, int] = (255, 255, 255),
        thickness: int = 3,
    ):
        """设置手部关键点的相关参数
        Args:
            radius: 关键点半径
            color: 关键点颜色
            thickness: 关键点粗细
        """
        # 绘制手部关键点参数
        self.point_radius: int = radius
        self.point_color: tuple[int, int, int] = color
        self.point_thickness: int = thickness

    def set_line(
        self,
        *,
        color: tuple[int, int, int] = (255, 255, 255),
        thickness: int = 3,
    ):
        """设置关键点连接线的相关参数
        Args:
            color: 关键点连接线的线条颜色
            thickness: 关键点连接线的粗细
        """
        # 绘制手部关键点连线参数
        self.line_color: tuple[int, int, int] = color
        self.line_thickness: int = thickness

    def draw_hand(
        self,
        image: np.ndarray | None = None,
        *,
        point_radius: int = 4,
        point_color: tuple[int, int, int] = (255, 255, 255),
        point_thickness: int = 3,
        line_color: tuple[int, int, int] = (255, 255, 255),
        line_thickness: int = 3,
    ) -> None | np.ndarray:
        """在原图中绘制手部,可同时设置手部绘制相关参数
        Args:
            image: 传入图片,将手部绘制到该图片上
            point_radius: 关键点半径
            point_color: 关键点颜色
            point_thickness: 关键点粗细
            line_color: 关键点连接线的线条颜色
            line_thickness: 关键点连接线的粗细
        """
        # 检查是否存在有帧图像
        if self.raw_img is None:
            return None
        # 设置绘制相关参数
        self.set_param(
            point_radius=point_radius,
            point_color=point_color,
            point_thickness=point_thickness,
            line_color=line_color,
            line_thickness=line_thickness,
        )
        # 默认是在原图里绘制
        output_img = self.raw_img
        if image is not None:
            output_img = image
        # 绘制
        x0, y0 = self.one_hand.raw_pos[0, :2].astype(np.int32)
        xp, yp = x0, y0
        for i in range(21):
            # 绘制关键点
            xi, yi = self.one_hand.raw_pos[i, :2].astype(np.int32)
            # 用颜色深度表示z轴
            point_z = self.one_hand.norm_pos[i, 2]
            z_color = tuple(
                map(
                    # lambda c: int((c / base_rgb) * (1 + (base_rgb - 1) * point_z)),
                    lambda c: int((c / 5) * (1 + 4 * point_z)),
                    self.point_color,
                )
            )
            cv2.circle(
                output_img, (xi, yi), self.point_radius, z_color, self.point_thickness
            )
            # 绘制手部连线
            cv2.line(output_img, (xp, yp), (xi, yi), z_color, self.line_thickness)
            if i % 4 == 0:
                xp, yp = x0, y0
            else:
                xp, yp = xi, yi
        return output_img

    def get_hand_img(self, *, padx: int = 0, pady: int = 0) -> None | np.ndarray:
        """截取对应名字的手部图片
        Args:
            padx: 横坐标方向的填充像素大小
            pady: 纵坐标方向的填充像素大小
        """
        # 检查是否存在有帧图像
        if self.raw_img is None:
            return None
        x0, y0, x1, y1 = self.one_hand.box  # 获取手部矩形框
        x0 -= padx  # 计算边缘
        x1 += padx
        y0 -= pady
        y1 += pady
        # 保证xxyy都为正整数
        x0, y0, x1, y1 = map(lambda x: x if x >= 0 else 0, (x0, y0, x1, y1))
        if x0 > x1 or y0 > y1:
            raise ValueError("It can't cut out a valid picture.")
        return self.raw_img[y0:y1, x0:x1, :].copy()

    def draw_hand_only(self, *, padx: int = 0, pady: int = 0) -> None | np.ndarray:
        """截取并绘制对应名字的手部图片
        Args:
            padx: 横坐标方向的填充像素大小
            pady: 纵坐标方向的填充像素大小
        """
        # 检查是否存在有图像
        hand_only_img = self.get_hand_img(padx=padx, pady=pady)
        if hand_only_img is None:
            return None
        # 统一边界大小
        hand_only_img = HandImage(hand_only_img, padx, pady)
        return self.draw_norm_hand(hand_only_img)

    def draw_norm_hand(self, bg_img: HandImage | np.ndarray) -> np.ndarray:
        """绘制归一化后的手部坐标
        Args:
            bg_img: 背景图片,需要传入HandImage实例
        """
        # 如果传入的是ndarray的话,需要转为HandImage类实例
        if isinstance(bg_img, np.ndarray):
            bg_img = HandImage(bg_img)
        x0, y0 = bg_img.calc_norm2img_pos(self.one_hand.norm_pos[0, :2])
        xp, yp = x0, y0
        for i in range(21):
            # 绘制关键点
            xi, yi = bg_img.calc_norm2img_pos(self.one_hand.norm_pos[i, :2])
            # 用颜色深度表示z轴
            point_z = self.one_hand.norm_pos[i, 2]
            z_color = tuple(
                map(
                    lambda c: int((c / 5) * (1 + 4 * point_z)),
                    self.point_color,
                )
            )
            cv2.circle(
                bg_img.data, (xi, yi), self.point_radius, z_color, self.point_thickness
            )
            # 绘制手部连线
            xi, yi = bg_img.calc_norm2img_pos(self.one_hand.norm_pos[i, :2])
            cv2.line(bg_img.data, (xp, yp), (xi, yi), z_color, self.point_thickness)
            if i % 4 == 0:
                xp, yp = x0, y0
            else:
                xp, yp = xi, yi
        return bg_img.data

    def draw_box(
        self,
        box_padx: int = 10,
        box_pady: int = 10,
        box_color: tuple[int, int, int] = (255, 0, 0),
        box_thickness: int = 3,
    ) -> None | np.ndarray:
        """在原图中绘制框住手部的矩形框
        Args:
            box_padx: 矩形框的横坐标方向的填充像素大小
            box_pady: 矩形框的纵坐标方向的填充像素大小
            box_color: 矩形框的线条颜色
            box_thickness: 矩形框的线条粗细
        """
        if self.raw_img is None:
            return None
        x0, y0, x1, y1 = self.one_hand.box
        cv2.rectangle(
            self.raw_img,
            (x0 - box_padx, y0 - box_pady),
            (x1 + box_padx, y1 + box_pady),
            box_color,
            box_thickness,
        )
        return self.raw_img


def draw_hand_data(
    data: np.ndarray,
    bg_img_w: int = 300,
    bg_img_h: int = 300,
    bg_colorBGR: tuple[int, int, int] = (0, 0, 0),
    padx: int = 30,
    pady: int = 30,
    point_radius: int = 4,
    point_colorBGR: tuple[int, int, int] = (255, 255, 255),
    point_thickness: int = 3,
    base_rgb: int = 4,
) -> np.ndarray:
    """根据一维手部数据绘制归一化后的手部坐标
    Args:
        data: 一维手部关键点数据
        bg_img_w: 纯色背景图片的宽
        bg_img_h: 纯色背景图片的高
        bg_colorBGR: 背景图片颜色
        padx: 横坐标方向的填充像素大小
        pady: 纵坐标方向的填充像素大小
        point_radius: 手部关键点的半径
        point_colorBGR: 手部关键点的颜色
        point_thickness: 手部关键点的粗细
        base_rgb: z轴基础颜色占比
    """
    # 创建纯色背景图
    bg_img = np.ones((bg_img_h, bg_img_w, 3), dtype=np.uint8)
    bg_img[:, :, :] = bg_colorBGR
    # data的前63列为手部数据的一维数组
    x0, y0 = data[:2]
    x0 = int(x0 * (bg_img_w - (2 * padx)) + padx)
    y0 = int(y0 * (bg_img_h - (2 * pady)) + pady)
    xp, yp = x0, y0
    for i in range(21):
        # 绘制关键点
        point_x, point_y, point_z = data[i * 3 : i * 3 + 3]
        point_x = int(point_x * (bg_img_w - (2 * padx)) + padx)
        point_y = int(point_y * (bg_img_h - (2 * pady)) + pady)
        # 用颜色深度表示z轴
        point_color = tuple(
            map(
                lambda c: int((c / base_rgb) * (1 + (base_rgb - 1) * point_z)),
                point_colorBGR,
            )
        )
        cv2.circle(
            bg_img, (point_x, point_y), point_radius, point_color, point_thickness
        )
        # 绘制手部连线
        xi, yi = data[i * 3 : i * 3 + 2]
        xi = int(xi * (bg_img_w - (2 * padx)) + padx)
        yi = int(yi * (bg_img_h - (2 * pady)) + pady)
        cv2.line(bg_img, (xp, yp), (xi, yi), point_color, point_thickness)
        if i % 4 == 0:
            xp, yp = x0, y0
        else:
            xp, yp = xi, yi
    return bg_img



from time import time

import numpy as np

from .HandInputScheme import HandInputScheme
from ..HandInput import HandInput


class FingerSwipeScheme(HandInputScheme):
    __slots__ = (
        "hand_input",
        "hand_name",
        "point_id",
        "swipe_velocity",
        "follow_velocity",
        "min_swipe_dist",
        "reset_time",
        "_follow_point",
        "_start_point",
        "_end_point",
        "_swipe_vec",
        "_swipe_dist",
        "_start_time",
    )

    def __init__(
        self,
        hand_input: HandInput,
        hand_name: str,
        point_id: int = 8,
        swipe_velocity: float = 3,
        follow_velocity: float = 1,
        min_swipe_dist: float = 0.3,
        reset_time: float = 2,
    ) -> None:
        """
        Args:
            hand_input: 手部输入类的实例
            hand_name: 手部名字,指定为哪只手制定手部操控方案
            point_id: 设置关键点索引作为滑动点,最好选指尖的点,默认为8(食指指尖)
            swipe_velocity: 手指滑动阈值,超过该阈值才设置跟随点,默认为3
            follow_velocity: 跟随点的速度,默认为1
            min_swipe_dist: 最小的滑动距离,超过该距离才看作滑动,默认为0.3
            reset_time: 重置跟随点的时间,超过该时间没有滑动则重置,默认为2秒
        """
        hand_input.hands_dict[hand_name]  # 获取该手部,没有该手部名字则报错
        self.hand_input: HandInput = hand_input
        self.hand_name: str = hand_name
        self.point_id: int = point_id
        self.swipe_velocity: float = swipe_velocity
        self.follow_velocity: float = follow_velocity
        self.min_swipe_dist: float = min_swipe_dist
        self.reset_time: float = reset_time
        # 创建跟随点,没有跟随设为None
        self._follow_point: np.ndarray | None = None
        # 记录滑动开始位置
        self._start_point: np.ndarray | None = None
        # 记录滑动结束位置
        self._end_point: np.ndarray = np.empty((1, 2))
        # 记录滑动向量
        self._swipe_vec: np.ndarray = np.empty((1, 2))
        # 记录滑动距离
        self._swipe_dist = 0
        # 记录开始跟随的时间,用于判断跟随点是否超时
        self._start_time = time()

    def _reset(self) -> None:
        """重置跟随点和计时器"""
        self._follow_point = None  # 重置跟随点为初始状态
        self._start_time = time()  # 重置计时器

    def update(self) -> None:
        """实时更新是否有滑动"""
        base = self.hand_input.base(self.hand_name)
        # 计算时间差查看是否超时
        if base is None or (time() - self._start_time) > self.reset_time:
            self._reset()  # 超时或没有检测到手部,则重置跟随点
            self._start_point = None  # 结束/关闭滑动的输出
            return None
        # 初始化跟随点
        if self._follow_point is None:
            self._follow_point = base.wrist_npos(self.point_id)[:2]
        # 更新跟随点的位置,跟随当前点的位置
        cur_point = base.wrist_npos(self.point_id)[:2]
        self._follow_point += (  # 更新跟随点位置
            (cur_point - self._follow_point)  # 与当前点位置距离差
            * self.follow_velocity  # 跟随点的速度
            * self.hand_input.delta_time  # 帧间间隔时间
        )
        # 超过速度阈值和跟随点距离阈值才返回输出
        if base.wrist_npos_velocity(self.point_id) > self.swipe_velocity:
            swipe_vec = cur_point - self._follow_point  # 计算滑动向量
            swipe_dist = np.linalg.norm(swipe_vec)  # 计算滑动距离
            # 检测滑动距离是否满足要求
            if swipe_dist > self.min_swipe_dist:
                self._start_point = self._follow_point.copy()  # 记录输出结果
                self._end_point = cur_point
                self._swipe_vec = swipe_vec
                self._swipe_dist = swipe_dist
                self._reset()  # 重置跟随点
                return None
        # 否则更新跟随点
        self._follow_point = base.wrist_npos(self.point_id)[:2]
        self._start_point = None  # 结束/关闭滑动的输出
        return None

    @property
    def is_activate(self) -> bool:
        """是否有滑动,有滑动则为True"""
        return self._start_point is not None

    @property
    def start_point(self) -> np.ndarray | None:
        """滑动开始位置坐标,如果没有滑动则返回None"""
        return self._start_point

    @property
    def end_point(self) -> np.ndarray | None:
        """滑动末尾位置坐标,如果没有滑动则返回None"""
        return self._end_point if self.is_activate else None

    @property
    def vector(self) -> np.ndarray | None:
        """返回滑动向量,如果没有滑动则返回None"""
        return self._swipe_vec if self.is_activate else None

    @property
    def distance(self) -> np.floating | float | None:
        """返回滑动的距离,如果没有滑动则返回None"""
        return self._swipe_dist if self.is_activate else None

    @property
    def norm_vec(self) -> np.ndarray | None:
        """返回滑动的方向向量,如果没有滑动则返回None"""
        return (self._swipe_vec / self._swipe_dist) if self.is_activate else None
